/*
* Copyright (c) 2003, University of Maryland 
* Copyright (c) 2003, Center for Satellite and Hybrid Communication Networks 
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the University of Maryland/Center for Satellite
*     and Hybrid Communication Networks nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

/* Authors: 
*     Jonathan Polley (jpolley@isr.umd.edu),
*     Dionysus Blazakis (dblaze@isr.umd.edu),
*     Jonathan McGee (mcgee@isr.umd.edu),
*     Daniel Rusk (ruskd@isr.umd.edu),
*     Karl Seamon (kks@isr.umd.edu),
*     Manish Karir (karir@isr.umd.edu)
*/

#include <assert.h>
#include <dlfcn.h>
#include <libxml/parser.h>

#include "defs.h"
#include "device.h"
#include "iomanager.h"
#include "debug.h"

error_t atmega_ioreg_write(atmega_chip *chip, uint16_t ioreg, uint8_t value){
    register_obj *reg;

    log_debug(chip, V_ALL, "write 0x%02x to ioreg 0x%02x\n", value, ioreg);

    reg=chip->register_head[ioreg];
    while(reg!=NULL){
        if(reg->write != NULL){
            reg->write(reg->dev, (value & reg->mask) >> reg->maskshift);
        }
        reg=reg->next;
    }

    return 0;
}

error_t atmega_ioreg_read(atmega_chip *chip, uint16_t ioreg, uint8_t *value){
    register_obj *reg;
    
    *value = chip->sram[ioreg];

    reg=chip->register_head[ioreg];
    while(reg!=NULL){
        if(reg->read != NULL){
            /* clear the bits set by this register */
            *value &= ~reg->mask;

            /* set them by reading */
            *value |= reg->read(reg->dev) << reg->maskshift;
        }
        reg=reg->next;
    }
    
    log_debug(chip, V_ALL, "read 0x%02x from ioreg 0x%02x\n", *value, ioreg);

    return 0;
}

int atmega_ioman_poll_interrupts(atmega_chip *chip){
    /* This function is called under the following assumptions:
     * - The CPU will call this function if interrupts are enabled.
     * - The CPU will immediately branch upon receiving an interrupt notification
     * - The individual devices handle specific interrupt masking
     * - This function will handle prioritizing, and will notify hardware to
     *   clear their interrupt flags
     * - This function should only be called if an interrupt is waiting (found
     *   by testing if chip->interrupts is non-zero)
     *
     * -1 is returned if no interrupt has occured (0 is a valid interrupt)
     * To convert the returned vector to an address, multiply by 2. */
    interrupt_obj *interrupt, *vectorbest = NULL;
    int bestint = 256;

    /* Make sure an interrupt is really waiting */
    assert(chip->interrupts != 0);

    /* Look for best interrupt (lowest vector number) */
    interrupt = chip->interrupt_head;
    while(interrupt != NULL) {
        if(interrupt->vector < bestint) {
            if(GET_FLAG(chip->interrupts, interrupt->vector_mask)) {
                bestint = interrupt->vector;
                vectorbest = interrupt;
            }
        }
        interrupt = interrupt->next;
    }

    /* If an interrupt is generated, clear it and notify the CPU */
    if(vectorbest != NULL) {
        CLR_FLAG(chip->interrupts, vectorbest->vector_mask);
        log_debug(chip, V_DEBUG, "Interrupt %d has been triggered, notifying processor.\n", bestint);
        vectorbest->clear(vectorbest->dev);
        return bestint;
    }

    return -1;
}

error_t atmega_ioman_tick_clocks(atmega_chip *chip){
    clock_obj *clock;
    
    clock = chip->clock_head;
    while(clock != NULL){
        if(clock->tick != NULL){
            clock->tick(clock->dev);
        }
        clock = clock->next;
    }

    return 0;
}

error_t atmega_ioman_enter_sleep(atmega_chip *chip, uint8_t sleepmode) {
    device_obj *device;
    static const char* sleepmodes[] = {
        "Idle", "ADC Noise Reduction", "Power-down", "Power-save",
        "Reserved", "Reserved", "Standby", "Extended Standby"
    };

    log_debug(chip, V_DEBUG, "CPU entering sleep mode '%s'\n", sleepmodes[sleepmode]);

    device = chip->device_head;
    while(device != NULL) {
        if(device->sleep != NULL) {
            device->sleep(device->dev, sleepmode);
        }
        device = device->next;
    }

    return 0;
}

error_t atmega_ioman_leave_sleep(atmega_chip *chip) {
    device_obj *device;

    log_debug(chip, V_DEBUG, "CPU exiting sleep\n");

    device = chip->device_head;
    while(device != NULL) {
        if(device->wake != NULL) {
            device->wake(device->dev);
        }
        device = device->next;
    }

    return 0;
}

error_t atmega_ioman_cleanup(atmega_chip *chip){
    device_obj *device;
    void *dlhandle;

    while(chip->device_head != NULL){
        /* free the stuff we allocated */
        xmlFree(chip->device_head->type);
        xmlFree(chip->device_head->name);
        dlhandle=chip->device_head->dlhandle;

        /* iterate and call destructor */
        device=chip->device_head;
        chip->device_head=chip->device_head->next;
        device->destroy(device->dev);

        dlclose(dlhandle);
    }

    return 0;
}

