/*
* Copyright (c) 2003, University of Maryland 
* Copyright (c) 2003, Center for Satellite and Hybrid Communication Networks 
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the University of Maryland/Center for Satellite
*     and Hybrid Communication Networks nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

/* Authors: 
*     Jonathan Polley (jpolley@isr.umd.edu),
*     Dionysus Blazakis (dblaze@isr.umd.edu),
*     Jonathan McGee (mcgee@isr.umd.edu),
*     Daniel Rusk (ruskd@isr.umd.edu),
*     Karl Seamon (kks@isr.umd.edu),
*     Manish Karir (karir@isr.umd.edu)
*/

/*
 *
 * ATmega128L Emulator
 *
 */


#include "defs.h"

#include <stdlib.h>
#include <signal.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <stdio.h>

#include "srec.h"
#include "atmega128.h"
#include "iomanager.h"
#include "debug.h"
#include "parse_config.h"
#include "table.h"

static char sreg_flags[9] = "CZNVSHTI";
static int interrupt=0;
static int currchip=0;
static int nchips=0;
static int backend = 0;
int verbosity=V_NORMAL;
uint64_t currentcycle = 0;

FILE *logfile=NULL;
int logstack=0;
int logmachine=0;
int logmemory=0;
int logcall=0;
int logintr=0;
int logtask=0;

int g_fd;

static void signal_handler(int type){
    signal(type, SIG_IGN);
    
    switch(type){
        case SIGINT:
            interrupt=1;
            break;
        case SIGCHLD:
            wait(NULL);
            break;
        default:
            printf("signal handler received unknown signal\n");
            break;
    }
    signal(type, signal_handler);
}

static void execute_chips(atmega_chip **chips, uint64_t currentcycle){
    int i;
    for(i=0; i<nchips; i++){
        /* execute current */
        if(chips[i]->cycles == currentcycle) {
            chips[i]->mode=ATEMU_MODE_EXECUTE;
            /* If time has been frozen, execute chip until no longer frozen */
            if(chips[i]->sram[0x00FA]) {
                uint64_t orgcycle = chips[i]->cycles;
                while(chips[i]->sram[0x00FA]) {
                    atmega_step(chips[i]);
                }
                chips[i]->cycles = orgcycle + 1;
            } else {
                atmega_step( chips[i] );
                /* break on unimplemented instructions */
                if(chips[i]->cycles == currentcycle){
                    interrupt = 1;
                }
            }
        }
        atmega_ioman_tick_clocks(chips[i]);
    }
}

static void decode_chips(atmega_chip **chips, uint64_t currentcycle){
    int i;
    
    for(i=0; i<nchips; i++){
        /* try and show where we will be, come next clock tick */
        if(chips[i]->am_sleeping){
            /* we are sleeping, assume we will continue */
            chips[i]->mode=ATEMU_MODE_DECODE;
            chips[i]->pc--;
            atmega_step( chips[i] );
            chips[i]->pc++;
        }else if(chips[i]->cycles == currentcycle){
            /* we broke right before the gicen instruction */
            chips[i]->mode=ATEMU_MODE_DECODE;
            atmega_step( chips[i] );
        }else{
            /* we are currently executing, and will continue to
             * execute this in the next cycle */
            uint32_t pc;
            uint64_t cycles;

            /* save values */
            pc=chips[i]->pc;
            cycles=chips[i]->cycles;

            /* move into the past */
            chips[i]->pc=chips[i]->oldpc;
            chips[i]->cycles=chips[i]->oldcycles;

            /* decode */
            chips[i]->mode=ATEMU_MODE_DECODE;
            atmega_step( chips[i] );

            /* restore values */
            chips[i]->pc=pc;
            chips[i]->cycles=cycles;
        }
    }
}

static int fork_compress (const char *fn)
{
    int fds[2], fd;
    int ires[2];
    unsigned int res;
    FILE *f;

    // hack -- better to discover problems before forking
    fd = open(fn, O_CREAT | O_WRONLY);
    if (fd == -1) {
        printf("couldn't open or create output file\n");
        exit(-1);
    }

    res = pipe (fds);
    if (res != 0) {
        printf("Output compress pipe failed.\n");
	return -1;
    }

    if (fork() != 0) {
        // parent -- close the write side and return the read side
        close (fds[0]);
	g_fd = fds[1];
	return fds[1];
    }

    // child -- copy the write side over STDIN, file to STDOUT
    ires[0] = dup2(fds[0], 0);
    ires[1] = dup2(fd, 1);
    if (ires[0] == -1 || ires[1] == -1) {
      printf("dup2 failed\n");
      exit (-1);
    }

    // make sure signal handlers are off for this child process
    signal(SIGINT, SIG_IGN);
    signal(SIGCHLD, SIG_IGN);
    close (fds[0]);
    close (fds[1]);

    // never returns
    res = execlp ("bzip2", "bzip2", "-9", NULL, NULL);

    printf("bzip2 exec failed\n");
    exit (-1);
}

static void dump( atmega_chip *chip, uint8_t *mem, uint16_t length, uint16_t begin )
{
    int addr, inchar;
    char memchar;

    mem += begin;

    for( addr = 0; addr < length; addr += 16 )
    {
        /* so I sort of cheat... */
        log_dump(chip, "");

        printf( "%06x:    ", addr + begin );
        printf( "%02x %02x %02x %02x %02x %02x %02x %02x  ",
            mem[addr], mem[addr + 1], mem[addr + 2], mem[addr + 3],
            mem[addr + 4], mem[addr + 5], mem[addr + 6], mem[addr + 7] );
        printf( "%02x %02x %02x %02x %02x %02x %02x %02x",
            mem[addr + 8], mem[addr + 9], mem[addr + 10], mem[addr + 11],
            mem[addr + 12], mem[addr + 13], mem[addr + 14], mem[addr + 15] );
        printf("    ");
        for(inchar = 0; inchar < 16; inchar++) {
            memchar = mem[addr + inchar];
            putchar(((memchar < 32) || (memchar > 126)) ? '.' : memchar);
        }
        printf("\n");
    }

    fflush(stdout);
}

static void unassemble( atmega_chip *chip, char *file ){
    uint32_t oldpc;
    int delta;
    int nnops=0;
    
    oldpc=chip->pc;
    chip->pc=0;
    chip->mode=ATEMU_MODE_DECODE;
    set_decode_mode(DECODE_DUMP, file);

    do{
        /* cheat around the fetch stage, to avoid having to print
         * thousands of NOPs at the end of the program */
        if(chip->flash[chip->pc]==0x0000){
            nnops++;
            delta=1;
            chip->pc++;
        }else{
            if(nnops>0){
                int i;
                chip->pc-=nnops;
                for(i=0; i<nnops; i++){
                    log_opcode(chip, "NOP\n");
                    chip->pc++;
                }
                nnops=0;
            }
            delta=atmega_step(chip);
            if(delta == -1){
                /* illegal instruction, do it as bytes */
                log_opcode(chip, ".byte\t0x%02x, 0x%02x\n", BOTTOM_BYTE(chip->flash[chip->pc]), TOP_BYTE(chip->flash[chip->pc]));
                delta=1;
            }
            chip->pc+=delta;
        }
    }while(chip->pc < chip->flash_words && (delta == 1 || delta == 2) && !interrupt);

    chip->pc=oldpc;
    set_decode_mode(DECODE_NORMAL, NULL);
}

static void dump_regs( atmega_chip *chip )
{
    char sreg[9];
    int i;
    int reg;

    for ( i = 0; i < 8; i++ )
        sreg[i] = ( GET_FLAG( *chip->sreg, 1 << i ) ) ? 
            (sreg_flags[i]) : (' ');
    sreg[8] = 0;

    printf( "[%lld] PC: 0x%08x\tSREG: 0x%02x[%s]\tSTACK: 0x%02x%02x\n",
        chip->cycles, chip->pc * 2, *chip->sreg, sreg, *chip->sph, *chip->spl );

    for( reg = 0; reg < 32; reg++ )
    {
        printf( "r%2d: %02x  ", reg, chip->register_file[reg] );

        if (! ( (reg + 1) % 8 ) )  printf( "\n" );
    }

    printf( "\n\n" );
}



static void dump_flash( atmega_chip *chip, uint32_t dump_addr , uint32_t dump_bytes)
{
    dump( chip, (uint8_t*)chip->flash, dump_bytes, dump_addr );
}

static void dump_sram( atmega_chip *chip, uint32_t dump_addr , uint32_t dump_bytes)
{
    dump( chip, (uint8_t*)chip->sram, dump_bytes, dump_addr );
}

static void has_changed_npoint( void *a, void *extra )
{
    watch_t *w = (watch_t *) a;
    atmega_chip *chip = (atmega_chip *) extra;
    uint8_t curr_val;

    atmega_read_sram(chip, w->addr, &(curr_val));

    if ( curr_val != w->value[0] )
    {
        w->value[0] = curr_val;

        log_debug(chip, V_QUIET,
                "notify addr %x id %d = %x, chip %d\n",
                w->addr, w->id, w->value[0], chip->id);
        fflush( stdout );
    }
}

static void mainloop_step(atmega_chip **chips) {
    int i;

    /* run current instruction */
    execute_chips(chips, currentcycle);
    currentcycle++;

    /* check notify points */
    for(i = 0; i < nchips; i++) {
        table_foreach(chips[i]->notifytable, has_changed_npoint, chips[i]);
    }

    for(i = 0; i < nchips; i++) {
        if (chips[i]->stack_changed != 0) {
            chips[i]->stack_changed = 0;
            stack_dump(chips[i]->pc, chips[i]->call_stack, chips[i]->id);
        }
    }

    /* decode next instruction */
    decode_chips(chips, currentcycle);
}

static void mainloop_go(atmega_chip **chips, int breakset, int watchset) {
    int i;
    int breakhit = 0; /* have we hit a brealpoint? */

    /* clear interrupt flag, if user pressed ctrl-c */
    interrupt = 0;

    do{
        execute_chips(chips, currentcycle);
        currentcycle++;

        if(currentcycle == halttime){
            interrupt = 1;
            printf("Execution halted due to time\n");
        }

        /* determine how many breakpoints have been hit */
        if(breakset > 0 || watchset > 0) {
            break_t *bpoint, query;
            for(i=0; i < nchips; i++) {
                if(chips[i]->cycles == currentcycle && !chips[i]->am_sleeping){
                    query.addr = chips[i]->pc;
                    bpoint = table_find(chips[i]->breaktable, &query);
                    if(bpoint != NULL){
                        breakhit++;
                        log_debug(chips[i], V_QUIET,
                                "breakpoint 0x%06x hit.\n", 2 * bpoint->addr);
                    }

                    /* trigger a "breakpoint" if a watch was hit */
                    if(chips[i]->watchhit != 0){
                        chips[i]->watchhit = 0;
                        breakhit++;
                    }
                }
            } /* for(i=0; i < nchips; i++) { */
        } /* if(breakset > 0 || watchset > 0) { */
    } while (!interrupt && !breakhit);

    decode_chips(chips, currentcycle);

    for(i=0; i < nchips; i++) {
        table_foreach( chips[i]->notifytable,
                has_changed_npoint, chips[i] );
        stack_dump( chips[i]->pc, chips[i]->call_stack, chips[i]->id );
    }
}

static void mainloop(atmega_chip **chips) {
    int quit = 0;
    /* number of current *points active */
    int breakset = 0;
    int watchset = 0;
    int i;
    char input[1024];
    /* variables utalized during parsing */
    char *parse, *lastchar;
    char **argv = NULL;
    int  argc, argv_alloc = 0;
    /* variables used in decoding parameters */
    int  intparse;
    uint32_t addrparse, addrbytes;
    break_t *bpoint, bquery;
    watch_t *wpoint, wquery;

    /* Make initial decodes upon starting program */
    for(i = 0; i < nchips; i++){
        chips[i]->mode = ATEMU_MODE_DECODE;
        atmega_step(chips[i]);
        stack_dump( chips[i]->pc, chips[i]->call_stack, chips[i]->id );
    }

    /* Prepare for parsing */
    argv_alloc = 8;
    argv = (char**)malloc(argv_alloc * sizeof(char*));

    while( !quit )
    {
        printf( "[atemu]('?' for help) " );
        if(backend){
            /* hackish...... */
            printf("\n");
        }
        fflush(stdout);

        if (fgets(input, sizeof(input), stdin)) 
        {
            /* Break command into argv/argc style */
            argc = 0;
            parse = strtok(input, " \t\n\r");
            while(parse != NULL) {
                /* Make sure we have space to store this argument */
                if(argv_alloc == argc) {
                    argv_alloc += 8;
                    argv = realloc(argv, argv_alloc * sizeof(char*));
                    if(argv == NULL) {
                        perror("realloc()");
                        exit(1);
                    }
                }

                /* Store this argument */
                argv[argc++] = parse;
                parse = strtok(NULL, " \t\n\r");
            }

            /* If argc is 0, a blank line was entered and we treat it as
             * a step */
            if(argc == 0) {
                mainloop_step(chips);
                continue;
            }

            /* Handle specified command */
            switch(toupper(argv[0][0])) {
            case 'Q':
                /* Quitting the program */
                quit = 1;
                break;

            case '?':
                /* Getting the command summary from the console */
                printf("Commands (case-insensitive):\n");
                printf("\tS or Enter \t\tStep one instruction\n");
                printf("\tD address [length]\tDump the SRAM\n");
                printf("\tF address [length]\tFlash dump\n");
                printf("\tR\t\t\tRegister dump\n");
                printf("\tU [output file]\t\tUnassemble program\n");
                printf("\tC chip number\t\tChip select\n");
                printf("\tB address\t\tBreakpoint at the Hex Address\n");
                printf("\tW address\t\tWatchpoint at the Hex Address\n");
                printf("\tN address\t\tNotfiypoint at the Hex Address\n"); 
                printf("\tX B|W|N addr\t\tDelete Break, Watch or Notify point at addr\n");
                printf("\tJ address\t\tJumps to the Hex Address\n");
                printf("\tG\t\t\tGo until breakpoint\n");
                printf("\tQ\t\t\tQuit\n");
                break;

            case 'C':
                /* Changing the active chip */

                /* Check arguments */
                if(argc < 2) {
                    printf("Error: The %c command requires a numerical"
                            " argument.\n", 'C');
                    continue;
                }

                /* Read the argument */
                intparse = strtoul(argv[1], &lastchar, 10);
                if((lastchar != NULL) && (*lastchar != 0)) {
                    /* Error: parameter contains illegal cahracters */
                    printf("Error: The value '%s' is not a valid"
                            " decimal number.\n", argv[1]);
                    continue;
                }

                /* Parameter doesn't contain any illegal characters */
                if((intparse < 0) || (intparse >= nchips)) {
                    /* Error: parameter out of range */
                    printf("Error: The value %d is out of range"
                            " (chips are numbered 0 through %d).\n",
                            intparse, nchips-1);
                    continue;
                }

                /* Change the current chip */
                currchip = intparse;
                printf("Current chip is now %d.\n", currchip);
                break;

            case 'S':
                /* Single stepping */
                mainloop_step(chips);
                break;

            case 'F':
                /* Dump bytes from flash */

                /* Check arguments */
                if(argc < 2) {
                    printf("Error: The %c command requires at least the"
                            " starting address to be provided.\n", 'F');
                    continue;
                }

                /* Parse the address */
                addrparse = strtoul(argv[1], &lastchar, 16);
                /* Round to the nearest 16 */
                addrparse = (addrparse / 16) * 16;

                /* Make sure it was a valid hexadecimal number */
                if((lastchar != NULL) && (*lastchar != 0)) {
                    /* Illegal character */
                    printf("Error: The value '%s' is not a valid"
                            " hexadecimal number.\n", argv[1]);
                    continue;
                }

                if(argc > 2) {
                    /* Parse the byte count */
                    addrbytes = strtoul(argv[2], &lastchar, 16);
                    /* Make sure it was a valid hexadecimal number */
                    if((lastchar != NULL) && (*lastchar != 0)) {
                        /* Illegal character */
                        printf("Error: The value '%s' is not a valid"
                                " hexadecimal number.\n", argv[2]);
                        continue;
                    }
                } else {
                    /* Default to 16 bytes if none provided */
                    addrbytes = 16;
                }

                /* Bound requested range to memory size */
                if(addrparse >= (chips[currchip]->flash_words * 2)) {
                    /* Beginning of range is after top of memory */
                    continue;
                }
                if((addrparse + addrbytes) >
                        (chips[currchip]->flash_words * 2)) {
                    /* End of range is after top of memory */
                    addrbytes = (chips[currchip]->flash_words * 2) - addrparse;
                }

                dump_flash(chips[currchip], addrparse, addrbytes);
                break;

            case 'J':
                /* Set's the program counter */

                /* Check arguments */
                if(argc < 2) {
                    printf("Error: The %c command requires a numerical"
                            " argument.\n", 'J');
                    continue;
                }

                /* Read the argument */
                addrparse = strtoul(argv[1], &lastchar, 16);
                if((lastchar != NULL) && (*lastchar != 0)) {
                    /* Error: parameter contains illegal cahracters */
                    printf("Error: The value '%s' is not a valid"
                            " hexadecimal number.\n", argv[1]);
                    continue;
                }

                /* Make sure the address is in range */
                addrparse /= 2;
                if(addrparse >= chips[currchip]->flash_words) {
                    printf("Error: The address '%s' is out of range"
                            " (addresses limited to 0 - 0x%X).\n",
                            argv[1], chips[currchip]->flash_words * 2 - 1);
                    continue;
                }

                chips[currchip]->pc = addrparse;
                chips[currchip]->mode = ATEMU_MODE_DECODE;
                atmega_step(chips[currchip]);
                break;

            case 'P':
                /* Poking holes in SRAM */

                /* Check arguments */
                if(argc < 3) {
                    printf("Error: The P command requires both an address"
                            " and a byte value to be provided.\n");
                    continue;
                }

                /* Parse the arguments */
                addrparse = strtoul(argv[1], &lastchar, 16);
                if((lastchar != NULL) && (*lastchar != 0)) {
                    printf("Error: The value '%s' is not a valid"
                            " hexadecimal number.\n", argv[1]);
                    continue;
                }
                addrbytes = strtoul(argv[2], &lastchar, 0);
                if((lastchar != NULL) && (*lastchar != 0)) {
                    printf("Error: The value '%s' is not a valid"
                            " numerical argument.\n", argv[2]);
                    continue;
                }

                /* Make sure the byte value is in range */
                if((addrbytes < -128) || (addrbytes > 255)) {
                    printf("Error: The value '%s' is larger than"
                            " one byte.\n", argv[2]);
                    continue;
                }

                /* Make sure the address is within range */
                if(addrparse >= chips[currchip]->sram_bytes) {
                    printf("Error: The address '%s' is out of range"
                            " (addresses limited to 0 - 0x%X).\n",
                            argv[1], chips[currchip]->sram_bytes - 1);
                    continue;
                }

                chips[currchip]->sram[addrparse] = addrbytes;
                break;

            case 'R':
                /* Dump the CPU registers in a human-readable manner */
                dump_regs(chips[currchip]);
                break;

            case 'D':
                /* Dump bytes from SRAM */

                /* Check arguments */
                if(argc < 2) {
                    printf("Error: The %c command requires at least the"
                            " starting address to be provided.\n", 'D');
                    continue;
                }

                /* Parse the address */
                addrparse = strtoul(argv[1], &lastchar, 16);
                /* Round to the nearest 16 */
                addrparse = (addrparse / 16) * 16;

                /* Make sure it was a valid hexadecimal number */
                if((lastchar != NULL) && (*lastchar != 0)) {
                    /* Illegal character */
                    printf("Error: The value '%s' is not a valid"
                            " hexadecimal number.\n", argv[1]);
                    continue;
                }

                if(argc > 2) {
                    /* Parse the byte count */
                    addrbytes = strtoul(argv[2], &lastchar, 16);
                    /* Make sure it was a valid hexadecimal number */
                    if((lastchar != NULL) && (*lastchar != 0)) {
                        /* Illegal character */
                        printf("Error: The value '%s' is not a valid"
                                " hexadecimal number.\n", argv[2]);
                        continue;
                    }
                } else {
                    /* Default to 16 bytes if none provided */
                    addrbytes = 16;
                }

                /* Bound requested range to memory size */
                if(addrparse >= chips[currchip]->sram_bytes) {
                    /* Beginning of range is after top of memory */
                    continue;
                }
                if((addrparse + addrbytes) > chips[currchip]->sram_bytes) {
                    /* End of range is after top of memory */
                    addrbytes = chips[currchip]->sram_bytes - addrparse;
                }

                dump_sram(chips[currchip], addrparse, addrbytes);
                break;

            case 'U':
                /* Unassemble the flash and dump to a file */
                if(argc < 2){
                    unassemble(chips[currchip], NULL);
                } else {
                    unassemble(chips[currchip], argv[1]);
                }
                break;

            case 'B':
                /* Create a breakpoint */

                /* Check parameters */
                if(argc < 2) {
                    printf("Error: The %c command requires a numerical"
                            " argument.\n", 'B');
                    continue;
                }

                /* Read the argument */
                addrparse = strtoul(argv[1], &lastchar, 16);
                if((lastchar != NULL) && (*lastchar != 0)) {
                    /* Error: parameter contains illegal cahracters */
                    printf("Error: The value '%s' is not a valid"
                            " hexadecimal number.\n", argv[1]);
                    continue;
                }

                /* Create the actual breakpoint */
                bpoint = (break_t*)malloc(sizeof(*bpoint));
                breakset++;
                bpoint->addr = addrparse / 2;
                log_debug(chips[currchip], V_QUIET,
                        "breakpoint 0x%06x set.\n", addrparse);
                table_store(chips[currchip]->breaktable, bpoint);
                break;

            case 'W':
                /* Create a watchpoint */

                /* Check parameters */
                if(argc < 2) {
                    printf("Error: The %c command requires a numerical"
                            " argument.\n", 'W');
                    continue;
                }

                /* Read the argument */
                addrparse = strtoul(argv[1], &lastchar, 16);
                if((lastchar != NULL) && (*lastchar != 0)) {
                    /* Error: parameter contains illegal cahracters */
                    printf("Error: The value '%s' is not a valid"
                            " hexadecimal number.\n", argv[1]);
                    continue;
                }

                if(argc == 3){
                    addrbytes = strtoul(argv[2], &lastchar, 16);
                    if((lastchar != NULL) && (*lastchar != 0)) {
                        /* Error: parameter contains illegal cahracters */
                        printf("Error: The value '%s' is not a valid"
                                " hexadecimal number.\n", argv[1]);
                        continue;
                    }
                    if(addrbytes == 0 || addrparse + addrbytes > chips[currchip]->sram_bytes){
                        printf("Error: Watchpoint size must be >0 and fit within SRAM\n");
                        continue;
                    }
                }else{
                    addrbytes = 1;
                }

                /* Create the actual watchpoint */
                wpoint = (watch_t*)malloc(sizeof(*wpoint));
                watchset++;
                wpoint->addr = addrparse;
                wpoint->length = addrbytes;
                wpoint->value = malloc(addrbytes);
                log_debug(chips[currchip], V_QUIET,
                        "watchpoint 0x%06x set.\n", addrparse);
                for(i=0; i < addrbytes; i++){
                    atmega_read_sram(chips[currchip], wpoint->addr + i, wpoint->value + i);
                }
                table_store(chips[currchip]->watchtable, wpoint);
                break;

            case 'N':
                /* Create a notify point */

                /* Check parameters */
                if(argc < 3) {
                    printf("Error: The %c command requires two numerical"
                            " arguments.\n", 'N');
                    continue;
                }

                /* Read the notify address */
                addrparse = strtoul(argv[1], &lastchar, 16);
                if((lastchar != NULL) && (*lastchar != 0)) {
                    /* Error: parameter contains illegal cahracters */
                    printf("Error: The value '%s' is not a valid"
                            " hexadecimal number.\n", argv[1]);
                    continue;
                }

                /* Read the notify identifier */
                intparse = strtol(argv[2], &lastchar, 0);
                if((lastchar != NULL) && (*lastchar != 0)) {
                    printf("Error: The value '%s' is not a valid"
                            " numerical value.\n", argv[2]);
                    continue;
                }

                /* Create the actual notifypoint */
                wpoint = (watch_t*)malloc(sizeof(*wpoint));
                wpoint->addr = addrparse;
                wpoint->id = intparse;
                wpoint->length = 1;
                wpoint->value = malloc(1);
                atmega_read_sram(chips[currchip], wpoint->addr,
                        wpoint->value);
                table_store(chips[currchip]->notifytable, wpoint);
                break;

            case 'X':
                /* Delete a (break|watch|notify) point */

                /* Check parameters */
                if(argc < 3) {
                    printf("Error: The %c command requires a break point type"
                            " and an address.\n", 'X');
                    continue;
                }

                /* Parse the address */
                addrparse = strtoul(argv[2], &lastchar, 16);
                if((lastchar != NULL) && (*lastchar != 0)) {
                    /* Error: parameter contains illegal cahracters */
                    printf("Error: The value '%s' is not a valid"
                            " hexadecimal number.\n", argv[2]);
                    continue;
                }

                /* Handle the appropriate type of breakpoint */
                switch(toupper(argv[1][0])) {
                case 'B':
                    /* Break Point */
                    bquery.addr = addrparse / 2;
                    if(table_delete(chips[currchip]->breaktable, &bquery,
                                free) >= 0) {
                        log_debug(chips[currchip], V_QUIET,
                                "breakpoint 0x%06x cleared.\n", addrparse);
                        breakset--;
                    }else{
                        log_debug(chips[currchip], V_QUIET,
                                "no such breakpoint.\n");
                    }
                    break;

                case 'W':
                    /* Watch Point */
                    wquery.addr = addrparse;
                    if(table_delete(chips[currchip]->watchtable, &wquery,
                                free) >= 0){
                        log_debug(chips[currchip], V_QUIET,
                                "watchpoint 0x%06x cleared.\n", addrparse);
                        watchset--;
                    }else{
                        log_debug(chips[currchip], V_QUIET,
                                "no such watchpoint.\n");
                    }
                    break;

                case 'N':
                    /* Notify Point */
                    wquery.addr = addrparse;
                    if(table_delete(chips[currchip]->notifytable, &wquery,
                                free) >= 0){
                        log_debug(chips[currchip], V_QUIET,
                                "notifypoint cleared.\n");
                    }else{
                        log_debug(chips[currchip], V_QUIET,
                                "no such notifypoint.\n");
                    }
                    
                    break;

                default:
                    printf("Error: Unknown breakpoint class '%s'.\n", argv[1]);
                    break;
                } /* switch(toupper(argv[1][0])) { */

                break;

            case 'G':
                /* Run until breakpoint */
                mainloop_go(chips, breakset, watchset);
                break;

            default:
                printf("Error: Unrecognized command '%s'\n", argv[0]);
            }
        } else { /* if ( fgets( input, sizeof(input), stdin) ) */
            if(feof(stdin)){
                quit = 1;
            }else{
                /* if they pressed ^C during fgets, forget it */
                interrupt = 0;
            }
        } /* if ( fgets( input, sizeof(input), stdin) ) */
    } /* while( !quit ) */

    free(argv);
    argv = NULL;
    argv_alloc = 0;
}

int main( int argc, char **argv )
{
    int cmderror = 0;
    error_t result;
    atmega_chip **chips;
    char *fname=NULL;
    char *config="mica2";
    char optstring[] = "htvqbc:l:VSTICMD";
    char option;
    int i, fd;

    while( (option = getopt(argc, argv, optstring))!=EOF){
        switch (option){
            case 'c':
                /* single-run: machine config */
                if(optarg){
                    config = optarg;
                }
                else{
                    printf("option c requires a file name\n");
                    cmderror=1;
                }
                break;
	    case 'l':
                if(optarg){
                    fd = fork_compress(optarg);
                    if (fd == -1) {
                        cmderror = 1;
                    } else {
                        logfile = fdopen(fd, "w");
                        if (logfile == NULL) {
                            perror("Error opening write pipe file");
                            cmderror = -1;
                        }
                    }
                } else {
                    printf("option l requires a file name\n");
                    cmderror = 1;
                }	        
	            break;
            case 'S':
                logstack = 1;
                break;
            case 'C':
                logcall = 1;
                break;
            case 'I':
                logintr = 1;
                break;
            case 'T':
                logtask = 1;
                break;
            case 'M':
                logmemory = 1;
                break;
            case 'D':
                logmachine = 1;
                srand48(time(NULL));
                break;
            case 'b':
                backend = 1;
                break;
            case 'h':
                printf("Options:\n");
                printf("\t-c <filename>\tspecify config file.  Default is mica2.conf\n");
                printf("\t-l <filename>\tspecify a log file.\n");
                printf("\t-C\t\tlog function calls, requires -l\n");
                printf("\t-S\t\tlog stack changes, requires -l\n");
                printf("\t-I\t\tlog interrupts and returns, requires -l\n");
                printf("\t-T\t\tlog TinyOS task info, requires -l\n");
                printf("\t-D\t\trandomly dump machine state info in log, requires -l\n");
                printf("\t-M\t\tlog memory operations, requires -l\n");
                printf("\t-t\t\trun in 'test mode'\n");
                printf("\t-b\t\trun in 'backend mode' (output altered for xatdb)\n");
                printf("\t-v\t\tincreases verbosity by one each time.\n");
                printf("\t-q\t\tdecreases verbosity by one each time.\n");
                printf("\t-V\t\tdisplay version information\n");
                printf("\t-h\t\tDisplay this list.\n\n");

                exit(0);
                break;
            case 'v':
                verbosity++;
                if(verbosity > V_ALL){
                    verbosity=V_ALL;
                }
                break;
            case 'q':
                verbosity--;
                if(verbosity < V_QUIET){
                    verbosity=V_QUIET;
                }
                break;
            case 'V':
                printf( "[ATmega128L Emulator %s]\n", PACKAGE_VERSION );
                exit(0);
                break;
            case ':':
                printf("missing required parameter for an option\n");
                cmderror=1;
                break;
            case '?':
                printf("unknown option\n");
                cmderror=1;
                break;
            default:
                printf("Unknown getopt return code '%c'\n", option);
                cmderror=1;
                break;
        }
    }

    if(optind < argc){
        fname = argv[optind];
    }
    
    if(cmderror || (fname == NULL)){
        printf("Usage: %s [-options] filename\n", argv[0]);
        printf("\tUse -h for option list\n");
        exit(1);
    }

    printf( "[ATmega128L Emulator %s]\n", PACKAGE_VERSION );

    /* set up the chips */
    nchips = atemu_load_file( &chips, fname, config );

    if ( nchips <= 0 ) {
        printf("Usage: %s [-options] filename\n", argv[0]);
        printf("\tUse -h for option list\n");
        exit(-1);
    }

    signal(SIGINT, signal_handler);
    signal(SIGCHLD, signal_handler);
    set_decode_mode(DECODE_NORMAL, NULL);

    mainloop( chips );

    for(i=0; i<nchips; i++){
        result = atmega_destroy_chip( &chips[i] );
        if ( result < 0 )
        {
            printf( "Error: Failed to destroy chip %d\n", i );
            return result;
        }
        printf( "Destroyed chip %d.\n", i ); 
    }

    if(logfile != NULL) {
        fclose(logfile);
        close(g_fd);
    }
    printf( "Done.\n" ); 

    return 0;
}
