/*
* Copyright (c) 2003, University of Maryland 
* Copyright (c) 2003, Center for Satellite and Hybrid Communication Networks 
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the University of Maryland/Center for Satellite
*     and Hybrid Communication Networks nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

/* Authors: 
*     Jonathan Polley (jpolley@isr.umd.edu),
*     Dionysus Blazakis (dblaze@isr.umd.edu),
*     Jonathan McGee (mcgee@isr.umd.edu),
*     Daniel Rusk (ruskd@isr.umd.edu),
*     Karl Seamon (kks@isr.umd.edu),
*     Manish Karir (karir@isr.umd.edu)
*/

#include "atmega128.h"
#include "opcodes.h"
#include "debug.h"

static char sreg_flags[]="CZNVSHTI";

error_t atmega_op_bset( atmega_chip *chip )
{
    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip, "SE%c\n", sreg_flags[chip->a]);
        return 1;
    }
    
    switch(chip->a){
        case 0:
            SET_FLAG(*chip->sreg, ATEMU_SREG_C);
            break;
        case 1:
            SET_FLAG(*chip->sreg, ATEMU_SREG_Z);
            break;
        case 2:
            SET_FLAG(*chip->sreg, ATEMU_SREG_N);
            break;
        case 3:
            SET_FLAG(*chip->sreg, ATEMU_SREG_V);
            break;
        case 4:
            SET_FLAG(*chip->sreg, ATEMU_SREG_S);
            break;
        case 5:
            SET_FLAG(*chip->sreg, ATEMU_SREG_H);
            break;
        case 6:
            SET_FLAG(*chip->sreg, ATEMU_SREG_T);
            break;
        case 7:
            SET_FLAG(*chip->sreg, ATEMU_SREG_I);
            break;
    }

    chip->cycles+=1;
    chip->pc++;

    return 0;
}

error_t atmega_op_bclr( atmega_chip *chip )
{
    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip, "CL%c\n", sreg_flags[chip->a]);
        return 1;
    }
    
    switch(chip->a){
        case 0:
            CLR_FLAG(*chip->sreg, ATEMU_SREG_C);
            break;
        case 1:
            CLR_FLAG(*chip->sreg, ATEMU_SREG_Z);
            break;
        case 2:
            CLR_FLAG(*chip->sreg, ATEMU_SREG_N);
            break;
        case 3:
            CLR_FLAG(*chip->sreg, ATEMU_SREG_V);
            break;
        case 4:
            CLR_FLAG(*chip->sreg, ATEMU_SREG_S);
            break;
        case 5:
            CLR_FLAG(*chip->sreg, ATEMU_SREG_H);
            break;
        case 6:
            CLR_FLAG(*chip->sreg, ATEMU_SREG_T);
            break;
        case 7:
            CLR_FLAG(*chip->sreg, ATEMU_SREG_I);
            break;
    }

    chip->cycles+=1;
    chip->pc++;

    return 0;
}

error_t atmega_op_bld( atmega_chip *chip )
{
    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip,  "BLD\tR%d, %d\n", chip->reg_index_a, chip->b );
        return 1;
    }

    if ( GET_FLAG( *chip->sreg, ATEMU_SREG_T ) )
    {
        SET_FLAG( chip->register_file[chip->reg_index_a], 1 << chip->b );
    }
    else
    {
        CLR_FLAG( chip->register_file[chip->reg_index_a], 1 << chip->b );
    }

    chip->cycles += 1;
    chip->pc++;

    return 0;
}

error_t atmega_op_bst( atmega_chip *chip )
{
    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip,  "BST\tR%d, %d\n", chip->reg_index_a, chip->b );
        return 1;
    }

    if ( GET_FLAG(chip->register_file[chip->reg_index_a], 1 << chip->b ) )
    {
        SET_FLAG( *chip->sreg, ATEMU_SREG_T );
    }
    else
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_T );
    }

    chip->cycles += 1;
    chip->pc++;

    return 0;
}

error_t atmega_op_sbi( atmega_chip *chip )
{
    uint8_t byte;

    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip,  "SBI\t0x%02x, %d\n", chip->a, chip->b );
        return 1;
    }

    atmega_read_sram( chip, chip->a + chip->io_offset, &byte );
    atmega_write_sram( chip, chip->a + chip->io_offset, byte | (1 << chip->b) );

    chip->cycles += 2;
    chip->pc++;    

    return 0;
}

error_t atmega_op_cbi( atmega_chip *chip )
{
    uint8_t byte;

    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip,  "CBI\t0x%02x, %d\n", chip->a, chip->b );
        return 1;
    }

    atmega_read_sram( chip, chip->a + chip->io_offset, &byte );
    atmega_write_sram( chip, chip->a + chip->io_offset, byte & (~(1 << chip->b)) );

    chip->cycles += 2;
    chip->pc++;

    return 0;
}

error_t atmega_op_swap( atmega_chip *chip )
{
    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip, "SWAP\tR%d\n", chip->reg_index_a);
        return 1;
    }

    chip->register_file[chip->reg_index_a] = (LOW_NIBBLE(chip->a) << 4) | HIGH_NIBBLE(chip->a);

    chip->pc += 1;
    chip->cycles += 1;

    return 0;
}

error_t atmega_op_asr( atmega_chip *chip )
{
    int n, c, v;
    
    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip, "ASR\tR%d\n", chip->reg_index_a);
        return 1;
    }

    if(chip->a & 0x01){
        SET_FLAG( *chip->sreg, ATEMU_SREG_C);
        c=1;
    }else{
        CLR_FLAG( *chip->sreg, ATEMU_SREG_C);
        c=0;
    }

    chip->a >>= 1;
    
    /* if former MSB is set (sign extend) */
    if(chip->a & 0x40){
        chip->a |= 0x80;
        SET_FLAG( *chip->sreg, ATEMU_SREG_N);
        n=1;
    }else{
        chip->a &= 0x7f;
        CLR_FLAG( *chip->sreg, ATEMU_SREG_N);
        n=0;
    }

    if(chip->a == 0){
        SET_FLAG( *chip->sreg, ATEMU_SREG_Z);
    }else{
        CLR_FLAG( *chip->sreg, ATEMU_SREG_Z);
    }

    if(n ^ c){
        SET_FLAG( *chip->sreg, ATEMU_SREG_V);
        v=1;
    }else{
        CLR_FLAG( *chip->sreg, ATEMU_SREG_V);
        v=0;
    }

    if(n ^ v){
        SET_FLAG( *chip->sreg, ATEMU_SREG_S);
    }else{
        CLR_FLAG( *chip->sreg, ATEMU_SREG_S);
    }
    
    chip->register_file[chip->reg_index_a] = chip->a;
    
    chip->pc += 1;
    chip->cycles += 1;
    
    return 0;
}

error_t atmega_op_lsr( atmega_chip *chip )
{
    int n, c, v;
    
    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip, "LSR\tR%d\n", chip->reg_index_a);
        return 1;
    }

    if(chip->a & 0x01){
        SET_FLAG( *chip->sreg, ATEMU_SREG_C);
        c=1;
    }else{
        CLR_FLAG( *chip->sreg, ATEMU_SREG_C);
        c=0;
    }

    chip->register_file[chip->reg_index_a] = (chip->a >> 1) & 0x7F;

    /* always clear top bit */
    CLR_FLAG( *chip->sreg, ATEMU_SREG_N);
    n=0;

    if(chip->register_file[chip->reg_index_a] == 0){
        SET_FLAG( *chip->sreg, ATEMU_SREG_Z);
    }else{
        CLR_FLAG( *chip->sreg, ATEMU_SREG_Z);
    }

    if(n ^ c){
        SET_FLAG( *chip->sreg, ATEMU_SREG_V);
        v=1;
    }else{
        CLR_FLAG( *chip->sreg, ATEMU_SREG_V);
        v=0;
    }

    if(n ^ v){
        SET_FLAG( *chip->sreg, ATEMU_SREG_S);
    }else{
        CLR_FLAG( *chip->sreg, ATEMU_SREG_S);
    }
    
    chip->pc += 1;
    chip->cycles += 1;
    
    return 0;
}

error_t atmega_op_ror( atmega_chip *chip )
{
    int n, c, v, newc;
    
    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip, "ROR\tR%d\n", chip->reg_index_a);
        return 1;
    }

    if(GET_FLAG(*chip->sreg, ATEMU_SREG_C)){
        c=1;
    }else{
        c=0;
    }

    if(chip->a & 0x01){
        SET_FLAG( *chip->sreg, ATEMU_SREG_C);
        newc = 1;
    }else{
        CLR_FLAG( *chip->sreg, ATEMU_SREG_C);
        newc = 0;
    }

    chip->register_file[chip->reg_index_a] = (chip->a >> 1) | (c << 7);

    /* if former carry flag is set */
    if(c){
        SET_FLAG( *chip->sreg, ATEMU_SREG_N);
        n=1;
    }else{
        CLR_FLAG( *chip->sreg, ATEMU_SREG_N);
        n=0;
    }

    if(chip->register_file[chip->reg_index_a] == 0){
        SET_FLAG( *chip->sreg, ATEMU_SREG_Z);
    }else{
        CLR_FLAG( *chip->sreg, ATEMU_SREG_Z);
    }

    if(n ^ newc){
        SET_FLAG( *chip->sreg, ATEMU_SREG_V);
        v=1;
    }else{
        CLR_FLAG( *chip->sreg, ATEMU_SREG_V);
        v=0;
    }

    if(n ^ v){
        SET_FLAG( *chip->sreg, ATEMU_SREG_S);
    }else{
        CLR_FLAG( *chip->sreg, ATEMU_SREG_S);
    }
    
    chip->pc += 1;
    chip->cycles += 1;
    
    return 0;
}
