/*
* Copyright (c) 2003, University of Maryland 
* Copyright (c) 2003, Center for Satellite and Hybrid Communication Networks 
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the University of Maryland/Center for Satellite
*     and Hybrid Communication Networks nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

/* Authors: 
*     Jonathan Polley (jpolley@isr.umd.edu),
*     Dionysus Blazakis (dblaze@isr.umd.edu),
*     Jonathan McGee (mcgee@isr.umd.edu),
*     Daniel Rusk (ruskd@isr.umd.edu),
*     Karl Seamon (kks@isr.umd.edu),
*     Manish Karir (karir@isr.umd.edu)
*/

#include "atmega128.h"
#include "opcodes.h"
#include "debug.h"

error_t atmega_op_ijmp( atmega_chip *chip )
{
    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip,  "IJMP\n");
        return 1;
    }

    chip->pc = PACK_INT_16(chip->register_file[ATEMU_REG_Z_HIGH],
            chip->register_file[ATEMU_REG_Z_LOW]);
    chip->cycles += 2;

    
    return 0;
}

/* EI* opcodes require a 22-bit PC */
error_t atmega_op_eijmp( atmega_chip *chip )
{
    uint8_t zl, zh, eind;

    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip,  "EIJMP\n");
        return 2;
    }

    zl = chip->register_file[ATEMU_REG_Z_LOW];
    zh = chip->register_file[ATEMU_REG_Z_HIGH];
    eind = *chip->eind;

    chip->pc = PACK_INT_24(eind, zh, zl);
    chip->cycles += 2;

    return 0;
}

error_t atmega_op_eicall( atmega_chip *chip )
{
    uint8_t zl, zh, eind;

    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip,  "EICALL\n");
        return 4;
    }

    atmega_push_help(chip, chip->pc+1);

    zl = chip->register_file[ATEMU_REG_Z_LOW];
    zh = chip->register_file[ATEMU_REG_Z_HIGH];
    eind = *chip->eind;

    if(logfile && logtask){
        fprintf(logfile, "%03d TASK %15ld\t%x\t%x\n", chip->id, (long)chip->cycles,
                (int)chip->pc, (int)PACK_INT_24(eind, zh, zl));
    }
    else if(logfile && logcall){
      fprintf(logfile, "%03d CALL %15ld\t%x\t%x\n", chip->id, (long)chip->cycles,
	      (int)chip->pc, (int)PACK_INT_24(eind, zh, zl));
    }

    chip->pc = PACK_INT_24(eind, zh, zl);
    chip->cycles += 4;

    return 0;
}

error_t atmega_op_icall( atmega_chip *chip )
{
    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip,  "ICALL\n");
        return 1;
    }

    if(logfile && logtask){
      fprintf(logfile, "%03d TASK %15ld\t%x\t%x\n", chip->id, (long)chip->cycles,
	      (int)chip->pc,  (int)PACK_INT_16(chip->register_file[ATEMU_REG_Z_HIGH],
					       chip->register_file[ATEMU_REG_Z_LOW]));
    }
    else if (logfile && logcall){
      fprintf(logfile, "%03d CALL %15ld\t%x\t%x\n", chip->id, (long)chip->cycles,
	      (int)chip->pc,  (int)PACK_INT_16(chip->register_file[ATEMU_REG_Z_HIGH],
					       chip->register_file[ATEMU_REG_Z_LOW]));
    }

    atmega_push_help( chip, chip->pc + 1 );

    chip->pc = PACK_INT_16(chip->register_file[ATEMU_REG_Z_HIGH],
            chip->register_file[ATEMU_REG_Z_LOW]);
    chip->cycles += 3;

    return 0;
}

error_t atmega_op_sbrc( atmega_chip *chip )
{
    uint8_t data;

    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip,  "SBRC\tR%d, %d\n", chip->reg_index_a, chip->b);
        return 1;
    }

    data = chip->register_file[chip->reg_index_a];

    if( GET_FLAG( data, 1 << chip->b ) == 0x00 ){
        /* skip next instruction */
        atmega_branch_skip_instruction( chip );
    }else{
        chip->pc += 1;
        chip->cycles += 1;
    }

    return 0;
}

error_t atmega_op_sbrs( atmega_chip *chip )
{
    uint8_t data;

    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip,  "SBRS\tR%d, %d\n", chip->reg_index_a, chip->b);
        return 1;
    }

    data = chip->register_file[chip->reg_index_a];

    if( GET_FLAG( data, 1 << chip->b ) ){
        /* skip next instruction */
        atmega_branch_skip_instruction( chip );
    }else{
        chip->pc += 1;
        chip->cycles += 1;
    }
 
    return 0;
}

error_t atmega_op_brbs( atmega_chip *chip )
{
    uint8_t branch = ( 1 << chip->b ) & (*chip->sreg);
    int8_t offset = ( 0x80 & (( chip->a & 0x40 ) << 1) ) | chip->a;

    if(chip->mode == ATEMU_MODE_DECODE){
        if((offset * 2) < 100){
            log_opcode(chip,  "BRBS\t%d, .%+d\t\t;0x%06x\n", chip->b, offset * 2, (chip->pc + offset + 1) * 2 );
        }else{
            log_opcode(chip,  "BRBS\t%d, .%+d\t;0x%06x\n", chip->b, offset * 2, (chip->pc + offset + 1) * 2 );
        }
        return 1;
    }

    if ( !branch )
    {
        chip->pc += 1;
        chip->cycles += 1;
    }
    else
    {
        chip->pc += offset + 1;
        chip->cycles += 2;
    }

    return 0;
}

error_t atmega_op_brbc( atmega_chip *chip )
{
    uint8_t not_branch = ( 1 << chip->b ) & (*chip->sreg);
    int8_t offset = ( 0x80 & (( chip->a & 0x40 ) << 1) ) | chip->a;

    if(chip->mode == ATEMU_MODE_DECODE){
        if((offset * 2) < 100){
            log_opcode(chip,  "BRBC\t%d, .%+d\t\t;0x%06x\n", chip->b, offset * 2, (chip->pc + offset + 1) * 2 );
        }else{
            log_opcode(chip,  "BRBC\t%d, .%+d\t;0x%06x\n", chip->b, offset * 2, (chip->pc + offset + 1) * 2 );
        }
        return 1;
    }

    if ( not_branch )
    {
        chip->pc += 1;
        chip->cycles += 1;
    }
    else
    {
        chip->pc += offset + 1;
        chip->cycles += 2;
    }

    return 0;
}

error_t atmega_op_rjmp( atmega_chip *chip )
{
    uint8_t top;
    int16_t s_addr;

    /* sign extend */

    top = (((HIGH_BYTE_16(chip->addr) & 0x08) >> 3) * 0xf);

    s_addr = (int16_t) ( PACK_INT_16( (HIGH_BYTE_16(chip->addr) | (top << 4) ),
        LOW_BYTE_16(chip->addr) ) );

    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip,  "RJMP\t.%+d\t\t;0x%06x\n", s_addr * 2, (chip->pc + s_addr + 1) * 2 );
        return 1;
    }

    chip->pc += s_addr + 1;
    chip->cycles += 2;

    return 0;
}

error_t atmega_op_rcall( atmega_chip *chip )
{
    uint8_t top;
    int16_t s_addr;

    /* sign extend */

    top = (((HIGH_BYTE_16(chip->addr) & 0x08) >> 3) * 0xf);

    s_addr = (int16_t) ( PACK_INT_16( (HIGH_BYTE_16(chip->addr) | (top << 4) ),
        LOW_BYTE_16(chip->addr) ) );

    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip,  "RCALL\t.%+d\t\t;0x%06x\n", s_addr * 2, (chip->pc + s_addr + 1) * 2 );
        return 1;
    }
    if(logfile && logcall){
      fprintf(logfile, "%03d RCALL %15ld\t%x\t%x\n", chip->id, (long)chip->cycles,
	      (int)chip->pc, (int)(chip->pc + s_addr + 1));
     }
    
    atmega_push_help( chip, chip->pc + 1 );

    chip->pc += s_addr + 1;
    chip->cycles += 3;

    return 0;
}

error_t atmega_op_sbis( atmega_chip *chip )
{
    uint8_t data;

    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip,  "SBIS\t0x%02x, %d\n", chip->a, chip->b);
        return 1;
    }

    atmega_read_sram( chip, chip->io_offset + chip->a, &data );

    if( GET_FLAG( data, 1<< chip->b ) ){
        /* skip next instruction */
        atmega_branch_skip_instruction( chip );
    }else{
        chip->pc += 1;
        chip->cycles += 1;
    }
 
    return 0;
}

error_t atmega_op_sbic( atmega_chip *chip )
{
    uint8_t data;

    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip,  "SBIC\t0x%02x, %d\n", chip->a, chip->b);
        return 1;
    }

    atmega_read_sram( chip, chip->io_offset + chip->a, &data );

    if( GET_FLAG( data, 1 << chip->b ) == 0x00 ){
        /* skip next instruction */
        atmega_branch_skip_instruction( chip );
    }else{
        chip->pc += 1;
        chip->cycles += 1;
    }
 
    return 0;
}

error_t atmega_op_jmp( atmega_chip *chip )
{
    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip,  "JMP\t0x%06x\n", chip->addr * 2 );
        return 2;
    }

    chip->pc = chip->addr;
    chip->cycles += 3;

    return 0;
}

error_t atmega_op_call( atmega_chip *chip )
{
    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip,  "CALL\t0x%06x\n", chip->addr * 2 );
        return 2;
    }
    if(logfile && logcall){
      fprintf(logfile, "%03d CALL %15ld\t%x\t%x\n", chip->id, (long)chip->cycles,
	      (int)chip->pc, (int)chip->addr);
    }

    atmega_push_help( chip, chip->pc + 2 );
    chip->pc = chip->addr;
    chip->cycles += 4;
 
    return 0;
}

error_t atmega_op_flags_jump_arith( atmega_chip *chip )
{
    error_t result;
    
    switch ( chip->bl_nibble )
    {
        case 0x0:
            /* COM */
            result = atmega_op_com(chip);
            break;
        case 0x1:
            /* NEG */
            result = atmega_op_neg(chip);
            break;
        case 0x2:
            /* SWAP */
            result = atmega_op_swap(chip);
            break;
        case 0x3:
            /* INC */
            result = atmega_op_inc(chip);
            break;
        case 0x5:
            /* ASR */
            result = atmega_op_asr(chip);
            break;
        case 0x6:
            /* LSR */
            result = atmega_op_lsr(chip);
            break;
        case 0x7:
            /* ROR */
            result = atmega_op_ror(chip);
            break;
        case 0x8:
            if(chip->tl_nibble == 0x4){
                /* set and clear */
                if(chip->bh_nibble & 0x8){
                    /* BCLR */
                    result = atmega_op_bclr(chip);
                }else{
                    /* BSET */
                    result = atmega_op_bset(chip);
                }
            }else{
                switch(chip->bh_nibble){
                    case 0x0:
                        /* RET */
                        result = atmega_op_ret(chip);
                        break;
                    case 0x1:
                        /* RETI */
                        result = atmega_op_reti(chip);
                        break;
                    case 0x8:
                        /* SLEEP */
                        result = atmega_op_sleep(chip);
                        break;
                    case 0x9:
                        /* BREAK */
                        result = atmega_op_break(chip);
                        break;
                    case 0xa:
                        /* WDR */
                        result = atmega_op_wdr(chip);
                        break;
                    case 0xc:
                        /* LPM */
                        result = atmega_op_lpm(chip);
                        break;
                    case 0xd:
                        /* ELPM */
                        result = atmega_op_elpm(chip);
                        break;
                    case 0xe:
                        /* SPM */
                        result = atmega_op_spm(chip);
                        break;
                    default:
                        log_error(chip, "unknown instruction\n");
                        result = -1;
                        break;
                }
            }
            break;
        case 0x9:
            /* a second lotta crap */
            if(chip->tl_nibble == 0x4){
                if(chip->bh_nibble == 0x0){
                    /* IJMP */
                    result = atmega_op_ijmp(chip);
                }else if(chip->bh_nibble == 0x1){
                    /* EIJMP */
                    result = atmega_op_eijmp(chip);
                }else{
                    log_error(chip, "unknown instruction\n");
                    result = -1;
                }
            }else if(chip->tl_nibble == 0x5){
                if(chip->bh_nibble == 0x0){
                    /* ICALL */
                    result = atmega_op_icall(chip);
                }else if(chip->bh_nibble == 0x1){
                    /* EICALL */
                    result = atmega_op_eicall(chip);
                }else{
                    log_error(chip, "unknown instruction\n");
                    result = -1;
                }
            }else{
                log_error(chip, "unknown instruction\n");
                result = -1;
            }

            break;
        case 0xa:
            /* DEC */
            result = atmega_op_dec( chip );
            break;
        case 0xc:
        case 0xd:
            result = atmega_op_jmp( chip );
            break;

        case 0xe:
        case 0xf:
            result = atmega_op_call( chip );
            break;

        default:
            log_error(chip,  "Unknown instruction\n" );
            result = -1;
    }

    return result;
}

error_t atmega_op_cpi( atmega_chip *chip )
{
    uint8_t value;

    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip,  "CPI\tR%d, 0x%02x\t;%d\n", chip->reg_index_a, chip->b, chip->b );
        return 1;
    }

    value = chip->a - chip->b;

    atmega_alu_set_sub_flags(chip, value, chip->a, chip->b);
    
    chip->cycles += 1;
    chip->pc++;

    return 0;
}

error_t atmega_op_cpse( atmega_chip *chip )
{
    /* this is an ugly instruction.
     * it skips one instruction, which may be one or two words.
     * so I have to partially deocde to test for each of the 4
     * two-word instructions.
     */

    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip,  "CPSE\tR%d, R%d\n", chip->reg_index_a, chip->reg_index_b );
        return 1;
    }

    if(chip->a == chip->b){
        /* skip next instruction */
        atmega_branch_skip_instruction( chip );
    }else{
        chip->pc += 1;
        chip->cycles += 1;
    }

    return 0;
}

error_t atmega_op_cp( atmega_chip *chip )
{
    uint8_t value;

    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip,  "CP\tR%d, R%d\n", chip->reg_index_a, chip->reg_index_b );
        return 1;
    }

    value = chip->a - chip->b;

    atmega_alu_set_sub_flags(chip, value, chip->a, chip->b);
    
    chip->cycles += 1;
    chip->pc++;

    return 0;
}

error_t atmega_op_cpc( atmega_chip *chip )
{
    uint8_t value, v, n;
    uint8_t mask_3 = 1 << 3;
    uint8_t mask_7 = 1 << 7;

    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip,  "CPC\tR%d, R%d\n", chip->reg_index_a, chip->reg_index_b );
        return 1;
    }

    value = chip->a - chip->b - ((*chip->sreg) & 0x1);

    if ( ( (~(chip->a & mask_3)) & (chip->b & mask_3) ) ||
         ( (chip->b & mask_3) & (value & mask_3) ) ||
         ( (value & mask_3) & (~(chip->a & mask_3)) ) )
    {
        SET_FLAG( *chip->sreg, ATEMU_SREG_H );
    }
    else
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_H );
    }

    if ( ( (chip->a & mask_7) & (~(chip->b & mask_7)) & (~(value & mask_7)) ) ||
         ( (~(chip->a & mask_7)) & (chip->b & mask_7) & (value & mask_7) ) )
    {
        SET_FLAG( *chip->sreg, ATEMU_SREG_V );
        v = 0x1;
    }
    else
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_V );
        v = 0x0;
    }

    if ( value & 0x80 )
    {
        SET_FLAG( *chip->sreg, ATEMU_SREG_N );
        n = 0x1;
    }
    else
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_N );
        n = 0x0;
    }

    if ( v ^ n )
    {
        SET_FLAG( *chip->sreg, ATEMU_SREG_S );
    }
    else
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_S );
    }

    if ( value != 0x00 )
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_Z );
    }

    if ( ( (~(chip->a & mask_7)) & (chip->b & mask_7) ) ||
         ( (chip->b & mask_7) & (value & mask_7) ) ||
         ( (value & mask_7) & (~(chip->a & mask_7)) ) )
    {
        SET_FLAG( *chip->sreg, ATEMU_SREG_C );
    }
    else
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_C );
    }

    chip->cycles += 1;
    chip->pc++;

    return 0;
}

error_t atmega_op_ret( atmega_chip *chip )
{
    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip,  "RET\n" );
        return 1;
    }
    if(logfile && logcall){
        fprintf(logfile, "%03d RETC %15ld\t%x\n", chip->id, (long)chip->cycles, (int)chip->pc);
    }
    
    atmega_pop_help( chip, &chip->pc );

    if ( chip->pc_bits == 16 )
        chip->cycles += 4;
    else if ( chip->pc_bits == 22 )
        chip->cycles += 5;

    return 0;
}

error_t atmega_op_reti( atmega_chip *chip )
{
    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip,  "RETI\n" );
        return 1;
    }
    if(logfile && logintr){
        fprintf(logfile, "%03d RETI %15ld\t%x\n", chip->id, (long)chip->cycles, (int)chip->pc);
    }
    
    atmega_pop_help( chip, &chip->pc );

    SET_FLAG( *chip->sreg, ATEMU_SREG_I );

    if ( chip->pc_bits == 16 )
        chip->cycles += 4;
    else if ( chip->pc_bits == 22 )
        chip->cycles += 5;

    return 0;
}
