/*
* Copyright (c) 2003, University of Maryland 
* Copyright (c) 2003, Center for Satellite and Hybrid Communication Networks 
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the University of Maryland/Center for Satellite
*     and Hybrid Communication Networks nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

/* Authors: 
*     Jonathan Polley (jpolley@isr.umd.edu),
*     Dionysus Blazakis (dblaze@isr.umd.edu),
*     Jonathan McGee (mcgee@isr.umd.edu),
*     Daniel Rusk (ruskd@isr.umd.edu),
*     Karl Seamon (kks@isr.umd.edu),
*     Manish Karir (karir@isr.umd.edu)
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include "srec.h"
#include "defs.h"

error_t srec_read(uint8_t *flash, int size, uint32_t *eaddr, char *srec){
    char buff[1024];
    FILE *sf;

    sf=fopen(srec, "r");

    if(sf==NULL){
        return -errno;
    }

    while(!feof(sf) && (fgets(buff, sizeof(buff), sf) != NULL)){
        int len, dlen=0;
        unsigned slen;
        char *data=NULL;
        uint32_t daddr;

        len=strlen(buff);
        if(buff[len-1]=='\n'){
            buff[len-1]='\0';
            len--;
        }
        if(buff[len-1]=='\r'){
            buff[len-1]='\0';
            len--;
        }
        
        if(buff[0]!='S'){
            fprintf(stderr, "\tError: Not an S-record\n");
            return -1;
        }
        
        sscanf(buff+2, "%2x", &slen);
        if(4+2*slen != len){
            fprintf(stderr, "\tWarning: length field does not match line"
                   " length: %d != %d\n", 4+2*slen, len);
        }
        
        switch(buff[1]){
            case '0':
                /* header record, unimportant */
                break;
            case '1':
                /* data, 2 byte address */
                if(len >= 4+4+2){
                    sscanf(buff+4, "%4x", &daddr);
                    data=buff+4+4;
                    dlen=strlen(data)-2;
                }else{
                    fprintf(stderr, "\tWarning: S1 record is too short\n");
                }
                break;
            case '2':
                /* data, 3 byte address */
                if(len >= 4+6+2){
                    sscanf(buff+4, "%6x", &daddr);
                    data=buff+4+6;
                    dlen=strlen(data)-2;
                }else{
                    fprintf(stderr, "\tWarning: S2 record is too short\n");
                }
                break;
            case '3':
                /* data, 4 byte address */
                if(len >= 4+8+2){
                    sscanf(buff+4, "%8x", &daddr);
                    data=buff+4+8;
                    dlen=strlen(data)-2;
                }else{
                    fprintf(stderr, "\tWarning: S3 record is too short\n");
                }
                break;
            case '5':
                /* count of prior records */
                /* don't need it, harddisks are reliable */
                break;
            case '7':
                /* starting execution address, 4 byte */
                if(len == 4+8+2){
                    sscanf(buff+4, "%8x", eaddr);
                }else{
                    fprintf(stderr, "\tWarning: S7 record is invalid length\n");
                }
                break;
            case '8':
                /* starting execution address, 3 byte */
                if(len == 4+6+2){
                    sscanf(buff+4, "%6x", eaddr);
                }else{
                    fprintf(stderr, "\tWarning: S8 record is invalid length\n");
                }
                break;
            case '9':
                /* starting execution address, 2 byte */
                if(len == 4+4+2){
                    sscanf(buff+4, "%4x", eaddr);
                }else{
                    fprintf(stderr, "\tWarning: S9 record is invalid length\n");
                }
                break;
            default:
                fprintf(stderr, "\tWarning: Invalid record type %c\n", buff[1]);
                break;
        }
        
        /* if there's data, and it fits, put it in */
        if(dlen>0 && ((daddr + dlen/2 - 1) < size)){
            uint32_t i;
            unsigned dbyte;

            for(i=0; i<dlen/2; i++){
                sscanf(data+2*i, "%2x", &dbyte);
                flash[daddr+i]=(uint8_t)dbyte;
            }

            dlen=0;
        }
    }

    return 0;
}

error_t srec_write(uint8_t *flash, int size, char *srec){
    return 0;
}
