/*
* Copyright (c) 2003, University of Maryland 
* Copyright (c) 2003, Center for Satellite and Hybrid Communication Networks 
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the University of Maryland/Center for Satellite
*     and Hybrid Communication Networks nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

/* Authors: 
*     Jonathan Polley (jpolley@isr.umd.edu),
*     Dionysus Blazakis (dblaze@isr.umd.edu),
*     Jonathan McGee (mcgee@isr.umd.edu),
*     Daniel Rusk (ruskd@isr.umd.edu),
*     Karl Seamon (kks@isr.umd.edu),
*     Manish Karir (karir@isr.umd.edu)
*/

#include <unistd.h>
#include <sys/types.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <gdk/gdk.h>
#include <assert.h>
#include <glib.h>

#include "atemu.h"
#include "chip_data.h"
#include "interface.h"
#include "gtk_helper.h"
#include "toolbar.h"
#include "debug.h"
#include "dialogs.h"
#include "support.h"
#include "gutter.h"
#include "chip_data_struct.h"

static FILE * pipe_out = NULL;
static int pipe_in_fd = -1;
static int pipe_out_fd = -1;
static GIOChannel *ioc;
static int chip_count;
static int status_bar_loading = FALSE;

/* is the chip currently in the 'go' state */
int xatdb_quit = 0;
int atemu_load_finished = 0;

pid_t child_pid = 0;
/* are we under the impression that we are running a child? */
static int child_alive = 0;
int iochannel_used = 0;

int pipe_write_safe = 0;
int chips_destroyed = 1;

/* called when data ready to be read is on fd. */
static gboolean atemu_console_process( GIOChannel *ioc, GIOCondition condition, gpointer data);

/************************************************************/
/* this is processes commands to be sent to the emulator.   */
/* it only sends them if it is safe (no deadlock) to do so. */
/************************************************************/

typedef void (atemu_idle_func)(void *);
static GSList *atemu_command_queue = NULL;
static GSList *atemu_pri_command_queue = NULL;

typedef struct atemu_command{
    atemu_idle_func *func;
    void *data;
}atemu_command;

/* this comparison is a little flaky, but all I care about is equality */
/* i.e. this is for finding, not sorting */
static int atemu_queue_find_func(gconstpointer a, gconstpointer b){
    atemu_command *ac = (atemu_command*)a, *bc = (atemu_command*)b;

    return ac->func != bc->func;
}

/* removes all queued calls to "func" from the queue */
static void atemu_queue_clear_func(atemu_idle_func *func){
    /* clear command X from the queue */
    GSList *found;
    atemu_command query;

    query.func = func;
    
    if(atemu_command_queue != NULL){
        verb_printf( FALSE, stdout, "clearing command <%p>\n", func);

        /* search through, remove instances */
        while((found = g_slist_find_custom(atemu_command_queue, &query, atemu_queue_find_func)) != NULL){
            atemu_command *cmd = found->data;

            atemu_command_queue = g_slist_remove(atemu_command_queue, cmd);
            free(cmd->data);
            free(cmd);
        }
    }
}

static void atemu_queue_clear_data(void *cmd, void *data){
    free(((atemu_command *)cmd)->data);
}

static void atemu_queue_clear(){
    /* clear commands */
    if(atemu_command_queue != NULL){
        verb_printf( FALSE, stdout, "clearing commands\n");

        g_slist_foreach(atemu_command_queue, atemu_queue_clear_data, NULL);
        g_slist_free(atemu_command_queue);
        atemu_command_queue = NULL;
    }
}

gboolean atemu_idle_process(void *data){
    atemu_command *cmd = NULL;

    /* see if the child has unexpectedly died */
    if(child_alive && child_pid == 0){
        child_error_popup();
        fclose(pipe_out);
        pipe_out = NULL;
        atemu_quit();
    }

    /* take from pri queue first, then normal */
    if(atemu_pri_command_queue != NULL){
        cmd = (atemu_command *)atemu_pri_command_queue->data;
        
        pipe_write_safe = 0;

        /* run the command */
        cmd->func(cmd->data);

        atemu_pri_command_queue = g_slist_remove(atemu_pri_command_queue, cmd);
        free(cmd);

        fflush( pipe_out );
    }else if(atemu_command_queue != NULL){
        if(pipe_out){
            if(pipe_write_safe){
                cmd = (atemu_command *)atemu_command_queue->data;
                /* some tasks process gtk idle loops */
                /* prevent two tasks from running at same time */
                pipe_write_safe = 0;
                
                /* run the command */
                cmd->func(cmd->data);

                atemu_command_queue = g_slist_remove(atemu_command_queue, cmd);
                free(cmd);

                fflush( pipe_out );
            }
        }else{
            /* there is no output pipe anymore, clear the normal commands */
            atemu_queue_clear();
        }
    }else{
        /* queue empty, sleep */
        usleep(1000);
    }
    
    return TRUE;
}

/* add (func, data) to queue */
void atemu_post_command(atemu_idle_func *func, void *data){
    atemu_command *cmd;

    cmd = (atemu_command*)malloc(sizeof(atemu_command));
    cmd->func = func;
    cmd->data = data;

    atemu_command_queue = g_slist_append(atemu_command_queue, cmd);
}

/* sticks important things on more important queue */
void atemu_post_pri_command(atemu_idle_func *func, void *data){
    atemu_command *cmd;

    cmd = (atemu_command*)malloc(sizeof(atemu_command));
    cmd->func = func;
    cmd->data = data;

    atemu_pri_command_queue = g_slist_append(atemu_pri_command_queue, cmd);
}

/********************************************/
/* these functions are global to all chips  */
/********************************************/

/* single opcode execution of the emulator */
static void atemu_step_idle(void *data){
    remove_pc_icon();
    fprintf( pipe_out, "step\n" );
}

void atemu_step()
{
    atemu_post_command(atemu_step_idle, NULL);
}

/* starts continuous execution of the emulator */
static void atemu_go_idle(void *data){
    remove_pc_icon();
    fprintf( pipe_out, "go\n" );
}

void atemu_go()
{
    atemu_post_command(atemu_go_idle, NULL);
}

/* interrupts execution of the emulator.  Does not quit it */
void atemu_stop()
{
    if ( child_pid > 0)
    {
        /* interrupt execution, if there is any */
        /* only stops the emulator's "go" command */
        kill( child_pid, SIGINT );

        /* need to remove any queued "go" or "step" type commands */
        /* note: this is an awkward hack */
        atemu_queue_clear_func(&atemu_go_idle);
        atemu_queue_clear_func(&atemu_step_idle);

        /* so as to help leave the program in a more-consistent state */
        /* read until the end of command token */
        while(!pipe_write_safe){
            atemu_console_process(ioc, G_IO_IN, NULL);
            g_main_context_iteration(NULL, FALSE);
        }

        /* prevent toolbar sensitivity race condition */
        toolbar_sense_stop();
    }
}

/* makes the child process exit */
static void atemu_quit_idle(void *data){
    /* all that other stuff was for the prior instance */
    atemu_queue_clear();
    
    if(pipe_out){
        if ( child_pid > 0){
            fprintf( pipe_out, "q\n" );
            fflush( pipe_out );
        }

        fclose(pipe_out);
        pipe_out = NULL;
    }

    /* A somewhat ugly hack.
     * We can't delete the data structures until the child pipe reader
     * callback is done writing to them.  so we wait until the child
     * is dead and the callback is deregistered.
     */
    while((child_pid > 0) || iochannel_used){
        g_main_context_iteration(NULL, FALSE);
        if(iochannel_used){
            // stupid main_iteration isn't reading from the pipe
            // so we do it manually here.
            atemu_console_process(ioc, G_IO_IN, NULL);
        }
    }

    /* clean up chips */
    if(chip_array_cnt > 0){
        destroy_chips();
    }
}

void atemu_quit()
{
    xatdb_quit = 1;
    child_alive = 0;
    atemu_stop();

    atemu_post_pri_command(atemu_quit_idle, NULL);
}

/* returns the status bar to a running state after loading.
 * this allows the full progress bar to display briefly before dissapearing */
static void atemu_status_bar_finish_idle( void *data ) 
{
    /* Return the status bar to its regular state */
    if( status_bar_loading ) {
        gtk_container_remove( GTK_CONTAINER( status_table ), statusbar1 );
        gtk_container_remove( GTK_CONTAINER( status_table ), progressbar1 );
        gtk_widget_set_size_request( progressbar1, 0, -1 );
        gtk_box_pack_end( GTK_BOX( statusbar1 ), progressbar1, FALSE, FALSE, 0 );
        gtk_table_attach_defaults( GTK_TABLE( status_table ), statusbar1, 0, 4, 0, 1 );
        gtk_statusbar_pop( GTK_STATUSBAR( statusbar1 ), 1 );
        gtk_progress_bar_set_fraction( GTK_PROGRESS_BAR( progressbar1 ), 0.0 );
        status_bar_loading = FALSE;
    }

    /* We did no useful work, but we still need to clear the "unsafe" flag */
    pipe_write_safe = 1;
}

void atemu_status_bar_finish(){
    atemu_post_pri_command( atemu_status_bar_finish_idle, NULL );
}

/* start an emulator with the given command line and environment */
/* returns < 0 on failure */
/* if you pass it in NULL for the filename, it tries to use the
 * same file it opened last time */
static void atemu_init_idle( void* data )
{
    int pipes_dtoe[2];
    int pipes_etod[2];
    static char * arg_file = NULL;
    char *file = (char*)data;
    GtkTextMark *mark;
    char *filename;
    GString *tooltip = g_string_new( "" );
    GString *helptip = g_string_new( "" );
    GString *loading_message = g_string_new( "" );

    /* should always have quit before we init again */
    assert(chips_destroyed);

    xatdb_quit = 0;
    atemu_load_finished = 0;

    pipe( pipes_dtoe );
    pipe( pipes_etod );

    /*close any lingering popups*/
    close_popups( );

    /* these strings are maintained for restart */
    if(file != NULL){
        if ( arg_file != NULL ){
            free( arg_file );
        }

        arg_file = strdup( file );
    }else{
        /* can't restart if we've never loaded a file */
        if(arg_file == NULL){
            return;
        }
    }

    /* Prepare status bar for loading feedback */
    
    if( !status_bar_loading ) {
        gtk_container_remove( GTK_CONTAINER( status_table ), statusbar1 );
        gtk_container_remove( GTK_CONTAINER( statusbar1 ), progressbar1 );
        gtk_widget_set_size_request( progressbar1, -1, -1 );
        gtk_table_attach_defaults( GTK_TABLE( status_table ), statusbar1, 0, 3, 0, 1 );
        gtk_table_attach_defaults( GTK_TABLE( status_table ), progressbar1, 3, 4, 0, 1 );
        status_bar_loading = TRUE;
    }
    gtk_progress_bar_set_fraction( GTK_PROGRESS_BAR( progressbar1 ), 0.0 );

    /* Initialize the chip counter */
    chip_count = 0;

    
    g_string_append( loading_message, _("Loading ") );
    g_string_append( loading_message, arg_file );
    
    gtk_statusbar_pop( GTK_STATUSBAR( statusbar1 ), 1 );
    gtk_statusbar_push( GTK_STATUSBAR( statusbar1 ), 1, loading_message->str );
    
    /* print feedback to the console */
    mark = text_buffer_printf( textbuff_console, "%s\n", loading_message->str );
    gtk_text_view_scroll_to_mark(GTK_TEXT_VIEW(textview_console), mark, 0.0, FALSE, 0.0, 0.0);

    g_string_free( loading_message, TRUE );

    child_pid = fork();
    child_alive = 1;

    if ( child_pid == 0 ) {
        char c;
        /* we have to do some crazy tricks sometimes... */
        char *munged_file;
        char saved_path[1024];

        close( 0 );
        close( 1 );
        close( 2 );
        dup( pipes_dtoe[0] );
        dup( pipes_etod[1] );
        dup( pipes_etod[1] );
        close( pipes_dtoe[0] );
        close( pipes_etod[1] );

        /* close parent's end of pipes */
        close( pipes_dtoe[1] );
        close( pipes_etod[0] );

        /* block on stdin so that the parent can be ready
         * for the child to start executing */
        read(0, &c, 1);

        /* now that the parent is ready, execute */
        /* try a few likely places, then $PATH */
        execl( "./atemu", "atemu", "-b", arg_file, NULL );

        /* we do a lot of magic to make it work if atemu is in .. */
        getcwd(saved_path, sizeof(saved_path));
        chdir("..");

        /* if it starts with ../ and atemu is in .., remove */
        /* this does not work in the general case, but it works in the
         * atemu src dir */
        if( strncmp(arg_file, "../", 3) == 0 ) {
            munged_file = arg_file + 3;
        }else{
            munged_file = arg_file;
        }
        execl( "./atemu", "atemu", "-b", munged_file, NULL );

        /* then just try the exe path, and standard relative pathnames */
        /* first, un-do the cd .. */
        chdir(saved_path);
        execlp( "atemu", "atemu", "-b", arg_file, NULL );

        printf( "Error: Could not start Atmega Emulator.\n" );
        exit(-1);
    }
    /* close child's end of pipes */
    close( pipes_dtoe[0] );
    close( pipes_etod[1] );

    pipe_out = fdopen( pipes_dtoe[1], "w" );
    pipe_out_fd = pipes_dtoe[1];
    if ( pipe_out == NULL ) 
    {
        verb_printf( TRUE, stdout, "[error] Could not open debug pipes.\n" );
        free(data);
        return;
    }

    pipe_in_fd = pipes_etod[0];

    ioc = g_io_channel_unix_new(pipe_in_fd);
    g_io_add_watch(ioc, G_IO_IN|G_IO_ERR, atemu_console_process, NULL);
    g_io_channel_set_close_on_unref(ioc, TRUE);
    iochannel_used++;

    /* tell that child to start (by writing one byte) */
    write(pipe_out_fd, "", 1);

    /* Set tooltips appropriate to filename */
    if( file != NULL ) { /* file is NULL when restarting */
        filename = strrchr( file, '/' );
        if ( filename == NULL )
        {
            /* in case the path lacks a '/' somehow */
            filename = file;
        }
        else
        {
            filename++;
        }
        
        /* Set tooltip for the Run button */
        g_string_append( tooltip, _("Run ") );
        g_string_append( tooltip, filename );
        g_string_append( helptip, _("Run ") );
        g_string_append( helptip, filename );
        g_string_append( helptip, _(" starting at the current line.") );
        gtk_tooltips_set_tip( GTK_TOOLTIPS( tooltips ), button_go, tooltip->str, helptip->str );
        
        /* Set tooltip for the Stop button */
        g_string_assign( tooltip, _("Stop ") );
        g_string_append( tooltip, filename );
        gtk_tooltips_set_tip( GTK_TOOLTIPS( tooltips ), button_stop, tooltip->str, tooltip->str );
        
        /* Set tooltip for the Restart button */
        g_string_assign( tooltip, _("Restart ") );
        g_string_append( tooltip, filename );
        gtk_tooltips_set_tip( GTK_TOOLTIPS( tooltips ), button_restart, tooltip->str, tooltip->str );
        
    }
    
    g_string_free( tooltip, TRUE );
    g_string_free( helptip, TRUE );
    free(data);
    
    return;
}

int atemu_init(const char *file){
    char *data = NULL;
    
    if(file != NULL){
        data = strdup(file);
    }
    atemu_post_pri_command(atemu_init_idle, data);
    
    return 0;
}

/* this gets queued up after the chip unassemblies to mark the end of loading */
static void atemu_load_finish_idle(void *data){
    GtkTextMark *mark;

    mark = text_buffer_printf( textbuff_console, _("...Finished Loading!\n") );
    gtk_text_view_scroll_to_mark(GTK_TEXT_VIEW(textview_console), mark, 0.0, FALSE, 0.0, 0.0);

    gtk_statusbar_pop( GTK_STATUSBAR( statusbar1 ), 1 );
    gtk_statusbar_push( GTK_STATUSBAR( statusbar1 ), 1, _("...Finished Loading!") );
    gtk_progress_bar_set_fraction( GTK_PROGRESS_BAR( progressbar1 ), 1.0 );
    gtk_widget_queue_draw( status_table );

    /* enqueue the status bar post load sequence */
    atemu_status_bar_finish();

    toolbar_sense_set(TRUE);

    atemu_load_finished = 1;

    /* We did no useful work, but we still need to clear the "unsafe" flag */
    pipe_write_safe = 1;
}

void atemu_load_finish(){
    atemu_post_command(atemu_load_finish_idle, NULL);
}

/********************************************/
/* these functions are specific to one chip */
/********************************************/

/* break/watch/etc points */
typedef struct breakpoint{
    uint32_t addr;
    uint32_t id;
    int size;
}breakpoint;

/* select chip */
static void atemu_chip_select_idle(void *data){
    int *chipid = (int*)data;
    fprintf( pipe_out, "c %d\n", *chipid );
    free(chipid);
}

/* add watchpoint */
static void atemu_add_watch_idle(void *data){
    breakpoint *point = (breakpoint *)data;
    fprintf( pipe_out, "w %x %x\n", point->addr, point->size );
    free(point);
}

void atemu_add_watch(int chipid, uint32_t addr, int size){
    breakpoint *point;
    int *chip;

    point = (breakpoint *)malloc(sizeof(point));
    point->addr = addr;
    point->size = size;
    
    chip = (int *)malloc(sizeof(int));
    *chip = chipid;

    atemu_post_command(atemu_chip_select_idle, chip);
    atemu_post_command(atemu_add_watch_idle, point);
}

/* add breakpoint */
static void atemu_add_break_idle(void *data){
    breakpoint *point = (breakpoint *)data;
    fprintf( pipe_out, "b %x\n", point->addr );
    free(point);
}

void atemu_add_break(int chipid, uint32_t addr){
    breakpoint *point;
    int *chip;

    point = (breakpoint *)malloc(sizeof(point));
    point->addr = addr;
    
    chip = (int *)malloc(sizeof(int));
    *chip = chipid;

    atemu_post_command(atemu_chip_select_idle, chip);
    atemu_post_command(atemu_add_break_idle, point);
}

/* add notifypoint */
static void atemu_add_notify_idle(void *data){
    breakpoint *point = (breakpoint *)data;
    fprintf( pipe_out, "n %x %d\n", point->addr, point->id );
    free(point);
}

void atemu_add_notify(int chipid, uint32_t addr, uint32_t id){
    breakpoint *point;
    int *chip;

    point = (breakpoint *)malloc(sizeof(breakpoint));
    point->addr = addr;
    point->id = id;
    
    chip = (int *)malloc(sizeof(int));
    *chip = chipid;

    atemu_post_command(atemu_chip_select_idle, chip);
    atemu_post_command(atemu_add_notify_idle, point);
}

/* clear watchpoint */
static void atemu_clear_watch_idle(void *data){
    breakpoint *point = (breakpoint *)data;
    fprintf( pipe_out, "x w %x\n", point->addr );
    free(point);
}

void atemu_clear_watch(int chipid, uint32_t addr){
    breakpoint *point;
    int *chip;

    point = (breakpoint *)malloc(sizeof(breakpoint));
    point->addr = addr;
    
    chip = (int *)malloc(sizeof(int));
    *chip = chipid;

    atemu_post_command(atemu_chip_select_idle, chip);
    atemu_post_command(atemu_clear_watch_idle, point);
}

/* clear breakpoint */
static void atemu_clear_break_idle(void *data){
    breakpoint *point = (breakpoint *)data;
    fprintf( pipe_out, "x b %x\n", point->addr );
    free(point);
}

void atemu_clear_break(int chipid, uint32_t addr){
    breakpoint *point;
    int *chip;

    point = (breakpoint *)malloc(sizeof(breakpoint));
    point->addr = addr;
    
    chip = (int *)malloc(sizeof(int));
    *chip = chipid;

    atemu_post_command(atemu_chip_select_idle, chip);
    atemu_post_command(atemu_clear_break_idle, point);
}

/* clear notifypoint */
static void atemu_clear_notify_idle(void *data){
    breakpoint *point = (breakpoint *)data;
    fprintf( pipe_out, "x n %x\n", point->addr );
    free(point);
}

void atemu_clear_notify(int chipid, uint32_t addr){
    breakpoint *point;
    int *chip;

    point = (breakpoint *)malloc(sizeof(breakpoint));
    point->addr = addr;
    
    chip = (int *)malloc(sizeof(int));
    *chip = chipid;

    atemu_post_command(atemu_chip_select_idle, chip);
    atemu_post_command(atemu_clear_notify_idle, point);
}

/* addresses for dumping */
/* base address, number of bytes */
typedef struct dumpaddr{
    uint32_t addr;
    uint32_t bytes;
}dumpaddr;

/* dumps register values */
static void atemu_dump_idle(void *data){
    dumpaddr *dump = (dumpaddr *)data;
    fprintf( pipe_out, "d %x %x\n", dump->addr, dump->bytes );
    free(dump);
}

void atemu_dump_regs(int chipid ){
    dumpaddr *dump;
    int *chip;

    chip = (int *)malloc(sizeof(int));
    *chip = chipid;
    atemu_post_command(atemu_chip_select_idle, chip);

    dump = (dumpaddr *)malloc(sizeof(dumpaddr));
    dump->addr = 0;
    dump->bytes = 32;
    atemu_post_command(atemu_dump_idle, dump);

    dump = (dumpaddr *)malloc(sizeof(dumpaddr));
    dump->addr = 0x50;
    dump->bytes = 16;
    atemu_post_command(atemu_dump_idle, dump);
}

/* dumps sram values */
void atemu_dump_sram(int chipid, long int addr ){
    dumpaddr *dump;
    int *chip;

    chip = (int *)malloc(sizeof(int));
    *chip = chipid;
    atemu_post_command(atemu_chip_select_idle, chip);

    dump = (dumpaddr *)malloc(sizeof(dumpaddr));
    dump->addr = addr;
    dump->bytes = 1;
    atemu_post_command(atemu_dump_idle, dump);
}

/* fetch disassembly */
static void atemu_unassemble_idle(void *data){
    GtkTextMark *mark;
    GString *fetch_message = g_string_new( "" );
    
    fprintf(pipe_out, "u\n");
    
    /* give the user feedback in the status bar */
    g_string_append_printf( fetch_message, _("Fetching disassembly for \"%s\" (%d of %d)..."),
        chip_array[chip_count].name, chip_count + 1, chip_array_cnt );
    
    gtk_statusbar_pop( GTK_STATUSBAR( statusbar1 ), 1 );
    gtk_statusbar_push( GTK_STATUSBAR( statusbar1 ), 1, fetch_message->str );
    
    if( chip_count == 0 ) {
        /* smaller increment since the disassembly loading is only starting now */
        gtk_progress_bar_set_fraction( GTK_PROGRESS_BAR( progressbar1 ), 
            gtk_progress_bar_get_fraction( GTK_PROGRESS_BAR( progressbar1 ) ) 
            + (double)( 1.0 / ( 1.0 + ( 2.0 * (double)chip_array_cnt ) ) ) );
    } else {
        /* a disassembly has completed, so use a larger increment */
        gtk_progress_bar_set_fraction( GTK_PROGRESS_BAR( progressbar1 ), 
            gtk_progress_bar_get_fraction( GTK_PROGRESS_BAR( progressbar1 ) ) 
            + (double)( 1.5 / ( 1.0 + ( 2.0 * (double)chip_array_cnt ) ) ) );
    }
    
    /* print feedback to the console */
    mark = text_buffer_printf( textbuff_console, "%s\n", fetch_message->str );
    gtk_text_view_scroll_to_mark(GTK_TEXT_VIEW(textview_console), mark, 0.0, FALSE, 0.0, 0.0);

    g_string_free( fetch_message, TRUE );

    chip_count++;
}

void atemu_unassemble(int chipid){
    int *chip;

    chip = (int *)malloc(sizeof(int));
    *chip = chipid;
    
    atemu_post_command(atemu_chip_select_idle, chip);
    atemu_post_command(atemu_unassemble_idle, NULL);
}

/* process output from atemu */
static gboolean atemu_console_process( GIOChannel *ioc, GIOCondition condition, gpointer data)
{
    char *buff;

    if(condition & G_IO_IN && g_io_channel_read_line(ioc, &buff, NULL, NULL, NULL) == G_IO_STATUS_NORMAL){
        process_buffer(buff);
        g_free(buff);
    }else{
        /* either EOF or an error */
        /* gtk will unref the ioc and close the file for us */
        iochannel_used = 0;
        return FALSE;
    }

    return TRUE;
}

